@@ -12,7 +12,7 @@
  *  warranty.  In no circumstances the author(s) shall be liable
  *  for any damages arising out of the use of this software.
  *
- *  $Id: addr.c,v 1.20 2004/08/02 12:20:54 shirok Exp $
+ *  $Id: addr.c,v 1.23 2006/01/22 00:52:54 shirok Exp $
  */
 
 #include "gauche/net.h"
@@ -128,11 +128,13 @@ static ScmObj sockaddr_un_allocate(ScmClass *klass, ScmObj initargs)
     memset(&addr->addr, 0, sizeof(struct sockaddr_un));
     addr->addr.sun_family = AF_UNIX;
     if (SCM_STRINGP(path)) {
-        int size = SCM_STRING_SIZE(path);
+        u_int size;
+        const char *cpath = Scm_GetStringContent(SCM_STRING(path), &size,
+                                                 NULL, NULL);
         if (size >= UNIX_ADDRESS_PATH_MAX-1) {
             Scm_Error("path too long: %S", path);
         }
-        memcpy(addr->addr.sun_path, SCM_STRING_START(path), size);
+        memcpy(addr->addr.sun_path, cpath, size);
         addr->addr.sun_path[size] = '\0';
     }
     addr->addrlen = sizeof(struct sockaddr_un);
@@ -263,15 +265,11 @@ void Scm_Init_NetAddr(ScmModule *mod)
     key_broadcast = SCM_MAKE_KEYWORD("broadcast");
     key_loopback  = SCM_MAKE_KEYWORD("loopback");
 
-    Scm_InitBuiltinClass(&Scm_SockAddrClass, "<sockaddr>", NULL,
-                         sizeof(ScmSockAddr), mod);
-    Scm_InitBuiltinClass(&Scm_SockAddrUnClass, "<sockaddr-un>", NULL,
-                         sizeof(ScmSockAddrUn), mod);
-    Scm_InitBuiltinClass(&Scm_SockAddrInClass, "<sockaddr-in>", NULL,
-                         sizeof(ScmSockAddrIn), mod);
+    Scm_InitStaticClass(&Scm_SockAddrClass, "<sockaddr>", mod, NULL, 0);
+    Scm_InitStaticClass(&Scm_SockAddrUnClass, "<sockaddr-un>", mod, NULL, 0);
+    Scm_InitStaticClass(&Scm_SockAddrInClass, "<sockaddr-in>", mod, NULL, 0);
 #ifdef HAVE_IPV6
-    Scm_InitBuiltinClass(&Scm_SockAddrIn6Class, "<sockaddr-in6>", NULL,
-                         sizeof(ScmSockAddrIn6), mod);
+    Scm_InitStaticClass(&Scm_SockAddrIn6Class, "<sockaddr-in6>", mod, NULL, 0);
 #endif /* HAVE_IPV6 */
 }
 
