@@ -313,6 +313,52 @@
 (uvcopy-startend-test "uvcopy-startend u64vector"
                       u64vector u64vector-copy u64vector-fill!)
 
+(define (uvcopy!-newapi-test msg make copy!)
+  (test* #`",msg /tstart" (make 0 7 8 9)
+         (copy! (make 0 1 2 3) 1 (make 7 8 9 10)))
+  (test* #`",msg /tstart(over)" (make 0 1 2 3)
+         (copy! (make 0 1 2 3) 4 (make 7 8 9 10)))
+  (test* #`",msg /tstart,,sstart" (make 0 9 10 3)
+         (copy! (make 0 1 2 3) 1 (make 7 8 9 10) 2))
+  (test* #`",msg /tstart,,sstart,,send" (make 0 1 2 9)
+         (copy! (make 0 1 2 3) 3 (make 7 8 9 10) 2 3))
+  )
+
+(uvcopy!-newapi-test "s8vector-copy! newapi" s8vector s8vector-copy!)
+(uvcopy!-newapi-test "u8vector-copy! newapi" u8vector u8vector-copy!)
+(uvcopy!-newapi-test "s16vector-copy! newapi" s16vector s16vector-copy!)
+(uvcopy!-newapi-test "u16vector-copy! newapi" u16vector u16vector-copy!)
+(uvcopy!-newapi-test "s32vector-copy! newapi" s32vector s32vector-copy!)
+(uvcopy!-newapi-test "u32vector-copy! newapi" u32vector u32vector-copy!)
+(uvcopy!-newapi-test "s64vector-copy! newapi" s64vector s64vector-copy!)
+(uvcopy!-newapi-test "u64vector-copy! newapi" u64vector u64vector-copy!)
+(uvcopy!-newapi-test "f32vector-copy! newapi" f32vector f32vector-copy!)
+(uvcopy!-newapi-test "f64vector-copy! newapi" f64vector f64vector-copy!)
+
+;;-------------------------------------------------------------------
+(test-section "swapping bytes")
+
+(test* "swapb s16"
+       '#s16(#x0123 #x4567 #x7654 #x3210)
+       (s16vector-swap-bytes '#s16(#x2301 #x6745 #x5476 #x1032)))
+(test* "swapb u16"
+       '#u16(#x0123 #x4567 #x7654 #x3210)
+       (u16vector-swap-bytes '#u16(#x2301 #x6745 #x5476 #x1032)))
+(test* "swapb s32"
+       '#s32(#x01234567 #x76543210)
+       (s32vector-swap-bytes '#s32(#x67452301 #x10325476)))
+(test* "swapb u32"
+       '#u32(#x01234567 #x76543210)
+       (u32vector-swap-bytes '#u32(#x67452301 #x10325476)))
+(test* "swapb s64"
+       '#s64(#x0123456789abcdef #x-123456789abcdf0)
+       (s64vector-swap-bytes
+        '#s64(#x-1032547698badcff #x1032547698badcfe)))
+(test* "swapb u64"
+       '#u64(#x0123456789abcdef #xfedcba9876543210)
+       (u64vector-swap-bytes
+        '#u64(#xefcdab8967452301 #x1032547698badcfe)))
+
 ;;-------------------------------------------------------------------
 (test-section "collection interface")
 
@@ -1143,6 +1189,28 @@
 (test* "string->u8vector (OOB)" *test-error*
        (string->u8vector "abcde" 2 6))
 
+(test* "string->u8vector!" '#u8(64 65 66 67 68)
+       (let ((v (u8vector 0 1 2 3 4)))
+         (string->u8vector! v 0 "@ABCD")))
+(test* "string->u8vector!" '#u8(64 65 66 67 68)
+       (let ((v (u8vector 0 1 2 3 4)))
+         (string->u8vector! v 0 "@ABCDEFGHIJKLMNOPQRTSUVWXYZ")))
+(test* "string->u8vector!" '#u8(64 65 66 3 4)
+       (let ((v (u8vector 0 1 2 3 4)))
+         (string->u8vector! v 0 "@AB")))
+(test* "string->u8vector!" '#u8(0 64 65 66 4)
+       (let ((v (u8vector 0 1 2 3 4)))
+         (string->u8vector! v 1 "@AB")))
+(test* "string->u8vector!" '#u8(0 1 2 3 65)
+       (let ((v (u8vector 0 1 2 3 4)))
+         (string->u8vector! v 4 "@ABCDE" 1)))
+(test* "string->u8vector!" '#u8(0 1 2 3 4)
+       (let ((v (u8vector 0 1 2 3 4)))
+         (string->u8vector! v 8 "@ABCDE" 1)))
+(test* "string->u8vector!" '#u8(0 1 2 3 4)
+       (let ((v (u8vector 0 1 2 3 4)))
+         (string->u8vector! v -1 "@ABCDE" 1)))
+
 (test* "u8vector->string" "@ABCD"
        (u8vector->string '#u8(64 65 66 67 68)))
 (test* "u8vector->string (start)" "ABCD"
@@ -1165,6 +1233,28 @@
 (test* "string->s8vector (OOB)" *test-error*
        (string->s8vector "abcde" 2 6))
 
+(test* "string->s8vector!" '#s8(64 65 66 67 68)
+       (let ((v (s8vector 0 1 2 3 4)))
+         (string->s8vector! v 0 "@ABCD")))
+(test* "string->s8vector!" '#s8(64 65 66 67 68)
+       (let ((v (s8vector 0 1 2 3 4)))
+         (string->s8vector! v 0 "@ABCDEFGHIJKLMNOPQRTSUVWXYZ")))
+(test* "string->s8vector!" '#s8(64 65 66 3 4)
+       (let ((v (s8vector 0 1 2 3 4)))
+         (string->s8vector! v 0 "@AB")))
+(test* "string->s8vector!" '#s8(0 64 65 66 4)
+       (let ((v (s8vector 0 1 2 3 4)))
+         (string->s8vector! v 1 "@AB")))
+(test* "string->s8vector!" '#s8(0 1 2 3 65)
+       (let ((v (s8vector 0 1 2 3 4)))
+         (string->s8vector! v 4 "@ABCDE" 1)))
+(test* "string->s8vector!" '#s8(0 1 2 3 4)
+       (let ((v (s8vector 0 1 2 3 4)))
+         (string->s8vector! v 8 "@ABCDE" 1)))
+(test* "string->s8vector!" '#s8(0 1 2 3 4)
+       (let ((v (s8vector 0 1 2 3 4)))
+         (string->s8vector! v -1 "@ABCDE" 1)))
+
 (test* "s8vector->string" "@ABCD"
        (s8vector->string '#s8(64 65 66 67 68)))
 (test* "s8vector->string (start)" "ABCD"
