@@ -1,7 +1,7 @@
 /*
  * load.c - load a program
  *
- *   Copyright (c) 2000-2004 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2005 Shiro Kawai, All rights reserved.
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: load.c,v 1.91 2004/07/28 05:13:22 shirok Exp $
+ *  $Id: load.c,v 1.106 2006/04/06 21:41:25 shirok Exp $
  */
 
 #include <stdlib.h>
@@ -82,19 +82,28 @@ static ScmObj key_paths              = SCM_UNBOUND;
 static ScmObj key_error_if_not_found = SCM_UNBOUND;
 static ScmObj key_environment        = SCM_UNBOUND;
 static ScmObj key_macro              = SCM_UNBOUND;
+static ScmObj key_ignore_coding      = SCM_UNBOUND;
 
 /*--------------------------------------------------------------------
  * Scm_LoadFromPort
  * 
  *   The most basic function in the load()-family.  Read an expression
  *   from the given port and evaluates it repeatedly, until it reaches
- *   EOF.  Then the port is closed.
+ *   EOF.  Then the port is closed.   The port is locked by the calling
+ *   thread until the operation terminates.
  *
  *   The result of the last evaluation remains on VM.
  *
  *   No matter how the load terminates, either normal or abnormal,
  *   the port is closed, and the current module is restored to the
  *   one when load is called.
+ *
+ *   FLAGS argument is ignored for now, but reserved for future
+ *   extension.  SCM_LOAD_QUIET_NOFILE and SCM_LOAD_IGNORE_CODING
+ *   won't have any effect for LoadFromPort; see Scm_Load below.
+ *
+ *   TODO: if we're using coding-aware port, how should we propagate
+ *   locking into the wrapped (original) port?
  */
 
 struct load_packet {
@@ -104,6 +113,7 @@ struct load_packet {
     ScmObj prev_port;
     ScmObj prev_history;
     ScmObj prev_next;
+    int    prev_situation;
 };
 
 /* Clean up */
@@ -111,11 +121,24 @@ static ScmObj load_after(ScmObj *args, int nargs, void *data)
 {
     struct load_packet *p = (struct load_packet *)data;
     ScmVM *vm = Scm_VM();
+
+#ifdef HAVE_GETTIMEOFDAY
+    if (SCM_VM_RUNTIME_FLAG_IS_SET(vm, SCM_COLLECT_LOAD_STATS)) {
+        struct timeval t0;
+        gettimeofday(&t0, NULL);
+        vm->stat.loadStat =
+            Scm_Cons(Scm_MakeIntegerU(t0.tv_sec*1000000+t0.tv_usec),
+                     vm->stat.loadStat);
+    }
+#endif /*HAVE_GETTIMEOFDAY*/
+
     Scm_ClosePort(p->port);
+    PORT_UNLOCK(p->port);
     Scm_SelectModule(p->prev_module);
     vm->load_port = p->prev_port;
     vm->load_history = p->prev_history;
     vm->load_next = p->prev_next;
+    vm->evalSituation = p->prev_situation;
     return SCM_UNDEFINED;
 }
 
@@ -127,7 +150,7 @@ static ScmObj load_cc(ScmObj result, void **data)
 
     if (!SCM_EOFP(expr)) {
         Scm_VMPushCC(load_cc, data, 1);
-        return Scm_VMEval(expr, SCM_UNBOUND);
+        return Scm_VMEval(expr, SCM_FALSE);
     } else {
         return SCM_TRUE;
     }
@@ -138,13 +161,15 @@ static ScmObj load_body(ScmObj *args, int nargs, void *data)
     return load_cc(SCM_NIL, &data);
 }
 
-ScmObj Scm_VMLoadFromPort(ScmPort *port, ScmObj next_paths, ScmObj env)
+ScmObj Scm_VMLoadFromPort(ScmPort *port, ScmObj next_paths,
+                          ScmObj env, int flags)
 {
     struct load_packet *p;
     ScmObj port_info;
     ScmVM *vm = Scm_VM();
     ScmModule *module = vm->module;
-    
+
+    /* Sanity check */
     if (!SCM_IPORTP(port))
         Scm_Error("input port required, but got: %S", port);
     if (SCM_PORT_CLOSED_P(port))
@@ -161,6 +186,7 @@ ScmObj Scm_VMLoadFromPort(ScmPort *port, ScmObj next_paths, ScmObj env)
     p->prev_port = vm->load_port;
     p->prev_history = vm->load_history;
     p->prev_next = vm->load_next;
+    p->prev_situation = vm->evalSituation;
 
     SCM_READ_CONTEXT_INIT(&(p->ctx));
     p->ctx.flags = SCM_READ_LITERAL_IMMUTABLE | SCM_READ_SOURCE_INFO;
@@ -171,6 +197,7 @@ ScmObj Scm_VMLoadFromPort(ScmPort *port, ScmObj next_paths, ScmObj env)
     vm->load_next = next_paths;
     vm->load_port = SCM_OBJ(port);
     vm->module = module;
+    vm->evalSituation = SCM_VM_LOADING;
     if (SCM_PORTP(p->prev_port)) {
         port_info = SCM_LIST2(p->prev_port,
                               Scm_MakeInteger(Scm_PortLine(SCM_PORT(p->prev_port))));
@@ -178,6 +205,8 @@ ScmObj Scm_VMLoadFromPort(ScmPort *port, ScmObj next_paths, ScmObj env)
         port_info = SCM_LIST1(SCM_FALSE);
     }
     vm->load_history = Scm_Cons(port_info, vm->load_history);
+
+    PORT_LOCK(port, vm);
     return Scm_VMDynamicWindC(NULL, load_body, load_after, p);
 }
 
@@ -186,13 +215,15 @@ static ScmObj load_from_port(ScmObj *args, int argc, void *data)
 {
     ScmPort *port;
     ScmObj paths, env;
+    int flags = 0;
+    
     if (!SCM_IPORTP(args[0])) {
         Scm_Error("input port required, but got %S", args[0]);
     }
     port = SCM_PORT(args[0]);
     paths = Scm_GetKeyword(key_paths, args[1], SCM_FALSE);
     env   = Scm_GetKeyword(key_environment, args[1], SCM_FALSE);
-    return Scm_VMLoadFromPort(port, paths, env);
+    return Scm_VMLoadFromPort(port, paths, env, flags);
 }
 
 static SCM_DEFINE_STRING_CONST(load_from_port_NAME, "load-from-port", 14, 14);
@@ -200,7 +231,7 @@ static SCM_DEFINE_SUBR(load_from_port_STUB, 1, 1,
                        SCM_OBJ(&load_from_port_NAME), load_from_port,
                        NULL, NULL);
 
-void Scm_LoadFromPort(ScmPort *port)
+void Scm_LoadFromPort(ScmPort *port, int flags)
 {
     Scm_Apply(SCM_OBJ(&load_from_port_STUB), SCM_LIST1(SCM_OBJ(port)));
 }
@@ -247,13 +278,13 @@ static ScmObj try_suffixes(ScmObj base, ScmObj suffixes)
 }
 
 ScmObj Scm_FindFile(ScmString *filename, ScmObj *paths,
-                    ScmObj suffixes, int error_if_not_found)
+                    ScmObj suffixes, int flags)
 {
-    int size = SCM_STRING_LENGTH(filename);
-    const char *ptr = SCM_STRING_START(filename);
+    u_int size;
+    const char *ptr = Scm_GetStringContent(filename, &size, NULL, NULL);
     int use_load_paths = TRUE;
     ScmObj file = SCM_OBJ(filename), fpath = SCM_FALSE;
-    
+
     if (size == 0) Scm_Error("bad filename to load: \"\"");
     if (*ptr == '~') {
         file = Scm_NormalizePathname(filename, SCM_PATH_EXPAND);
@@ -283,7 +314,7 @@ ScmObj Scm_FindFile(ScmString *filename, ScmObj *paths,
         if (SCM_PAIRP(lpath)) {
             *paths = SCM_CDR(lpath);
             return SCM_OBJ(fpath);
-        } else if (error_if_not_found) {
+        } else if (!(flags&SCM_LOAD_QUIET_NOFILE)) {
             Scm_Error("cannot find file %S in *load-path* %S", file, *paths);
         } else {
             *paths = SCM_NIL;
@@ -292,71 +323,109 @@ ScmObj Scm_FindFile(ScmString *filename, ScmObj *paths,
         *paths = SCM_NIL;
         fpath = try_suffixes(file, suffixes);
         if (!SCM_FALSEP(fpath)) return fpath;
-        if (error_if_not_found) {
+        if (!(flags&SCM_LOAD_QUIET_NOFILE)) {
             Scm_Error("cannot find file %S to load", file);
         }
     }
     return SCM_FALSE;
 }
 
-/*
- * Load
+/*---------------------------------------------------------------------
+ * Scm_Load
+ *
+ *  Scheme's load().
+ * 
+ *  filename   - name of the file.  can be sans suffix.
+ *  load_paths - list of pathnames or #f.   If #f, system's load path
+ *               is used.
+ *  env        - a module where the forms are evaluated, or #f.
+ *               If #f, the current module is used.
+ *  flags      - combination of bit flags
+ *               SCM_LOAD_QUIET_NOFILE, SCM_LOAD_IGNORE_CODING
  */
 
 ScmObj Scm_VMLoad(ScmString *filename, ScmObj load_paths,
-                  ScmObj env, int errorp)
+                  ScmObj env, int flags)
 {
     ScmObj port, truename, suffixes;
     ScmVM *vm = Scm_VM();
+    int errorp = !(flags&SCM_LOAD_QUIET_NOFILE);
+    int ignore_coding = flags&SCM_LOAD_IGNORE_CODING;
 
     suffixes = SCM_GLOC_GET(ldinfo.load_suffixes_rec);
     if (!SCM_PAIRP(load_paths)) load_paths = Scm_GetLoadPath();
-    truename = Scm_FindFile(filename, &load_paths, suffixes, errorp);
+    truename = Scm_FindFile(filename, &load_paths, suffixes, flags);
     if (SCM_FALSEP(truename)) return SCM_FALSE;
+
+#ifdef HAVE_GETTIMEOFDAY
+    if (SCM_VM_RUNTIME_FLAG_IS_SET(vm, SCM_COLLECT_LOAD_STATS)) {
+        struct timeval t0;
+        gettimeofday(&t0, NULL);
+        vm->stat.loadStat =
+            Scm_Acons(truename,
+                      Scm_MakeIntegerU(t0.tv_sec*1000000+t0.tv_usec),
+                      vm->stat.loadStat);
+    }
+#endif /*HAVE_GETTIMEOFDAY*/
     if (SCM_VM_RUNTIME_FLAG_IS_SET(vm, SCM_LOAD_VERBOSE)) {
         int len = Scm_Length(vm->load_history);
         SCM_PUTZ(";;", 2, SCM_CURERR);
         while (len-- > 0) SCM_PUTC(' ', SCM_CURERR);
         Scm_Printf(SCM_CURERR, "Loading %A...\n", truename);
     }
+
     port = Scm_OpenFilePort(Scm_GetStringConst(SCM_STRING(truename)),
                             O_RDONLY, SCM_PORT_BUFFER_FULL, 0);
     if (SCM_FALSEP(port)) {
         if (errorp) Scm_Error("file %S exists, but couldn't open.", truename);
         else        return SCM_FALSE;
     }
-    return Scm_VMLoadFromPort(SCM_PORT(port), load_paths, env);
+    if (!ignore_coding) {
+        port = Scm_MakeCodingAwarePort(SCM_PORT(port));
+    }
+    return Scm_VMLoadFromPort(SCM_PORT(port), load_paths, env, flags);
 }
 
-/* Scheme subr (%load filename &keyword paths error-if-not-found environment) */
+/* Scheme subr (%load filename &keyword paths error-if-not-found
+                                        environment aware-coding) */
 static ScmObj load(ScmObj *args, int argc, void *data)
 {
     ScmString *file;
     ScmObj paths, env;
-    int errorp;
+    int flags = 0;
+    
     if (!SCM_STRINGP(args[0])) {
         Scm_Error("string required, but got %S", args[0]);
     }
     file = SCM_STRING(args[0]);
-    paths  = Scm_GetKeyword(key_paths, args[1], SCM_FALSE);
-    errorp = !SCM_FALSEP(Scm_GetKeyword(key_error_if_not_found, args[1], SCM_TRUE));
-    env    = Scm_GetKeyword(key_environment, args[1], SCM_FALSE);
-    return Scm_VMLoad(file, paths, env, errorp);
+    paths = Scm_GetKeyword(key_paths, args[1], SCM_FALSE);
+    env   = Scm_GetKeyword(key_environment, args[1], SCM_FALSE);
+    if (SCM_FALSEP(Scm_GetKeyword(key_error_if_not_found, args[1], SCM_TRUE)))
+        flags |= SCM_LOAD_QUIET_NOFILE;
+    if (!SCM_FALSEP(Scm_GetKeyword(key_ignore_coding, args[1], SCM_FALSE)))
+        flags |= SCM_LOAD_IGNORE_CODING;
+    return Scm_VMLoad(file, paths, env, flags);
 }
 
 static SCM_DEFINE_STRING_CONST(load_NAME, "load", 4, 4);
 static SCM_DEFINE_SUBR(load_STUB, 1, 1, SCM_OBJ(&load_NAME), load, NULL, NULL);
 
 
-int Scm_Load(const char *cpath, int errorp)
+int Scm_Load(const char *cpath, int flags)
 {
     ScmObj r, f = SCM_MAKE_STR_COPYING(cpath);
-    if (errorp) {
-        r = Scm_Apply(SCM_OBJ(&load_STUB), SCM_LIST1(f));
-    } else {
-        r = Scm_Apply(SCM_OBJ(&load_STUB),
-                      SCM_LIST3(f, key_error_if_not_found, SCM_FALSE));
+    ScmObj options = SCM_NIL;
+    
+    if (flags&SCM_LOAD_QUIET_NOFILE) {
+        options = Scm_Cons(key_error_if_not_found,
+                           Scm_Cons(SCM_FALSE, options));
     }
+    if (flags&SCM_LOAD_IGNORE_CODING) {
+        options = Scm_Cons(key_ignore_coding,
+                           Scm_Cons(SCM_TRUE, options));
+    }
+    
+    r = Scm_Apply(SCM_OBJ(&load_STUB), Scm_Cons(f, options));
     return !SCM_FALSEP(r);
 }
 
@@ -456,6 +525,59 @@ ScmObj Scm_AddLoadPath(const char *cpath, int afterp)
     return r;
 }
 
+#if 0 /* not sure we need this yet */
+/*------------------------------------------------------------------
+ * Append FEATURE to *cond-features*.  Once added, the symbol FEATURE
+ * will be recognized by cond-expand.
+ *
+ *  (cond-expand (FEATURE body ...)) 
+ *   
+ * If loading a module is required in order to make FEATURE available,
+ * such module can be specified in MODULE argument.  It can be SCM_FALSE
+ * if the feature is built-in.
+ *
+ * Although cond-expand can be used like #ifdefs of cpp, it is a bit
+ * different since you can test a feature _before_ loading the module
+ * (or, you can say testing the feature itself implies loading the
+ * module if available).  Thus it is not recommended to add a feature
+ * as a result of loading a specific module.  One possible usage is
+ * to add a feature conditionally if a module may or may not provide
+ * such a feature depending on platforms or compile-time options.
+ * (An example of this is gauche.net, which may provide gauche-ipv6
+ * if IPv6 support is compiled in.)
+ *
+ * Because of this obscurity, I do not encourage user-land programs
+ * to use this API.  We may try some use within Gauche bundled extensions,
+ * and see how well it works, before we make it open to public.
+ * (Esp. we need to come up a reasonable naming scheme of features
+ * if we allow arbitrary extensions to add them).
+ */
+
+ScmObj
+Scm_AddCondFeature(ScmObj feature, ScmObj module)
+{
+    ScmObj cell, z;
+   
+    if (!SCM_SYMBOLP(feature)) {
+        Scm_Error("feature name must be a symbol, but got %S", feature);
+    }
+    if (!SCM_SYMBOLP(module) && !SCM_FALSEP(module)) {
+        Scm_Error("module must be a symbol or #f, but got %S", module);
+    }
+    if (SCM_FALSEP(module)) {
+        cell = SCM_LIST1(feature);
+    } else {
+        cell = SCM_LIST2(feature, module);
+    }
+    
+    (void)SCM_INTERNAL_MUTEX_LOCK(ldinfo.path_mutex);
+    z = ldinfo.cond_features_rec->value =
+        Scm_Cons(cell, ldinfo.cond_features_rec->value);
+    (void)SCM_INTERNAL_MUTEX_UNLOCK(ldinfo.path_mutex);
+    return z;
+}
+#endif
+
 /*------------------------------------------------------------------
  * Dynamic link
  */
@@ -472,7 +594,6 @@ ScmObj Scm_AddLoadPath(const char *cpath, int afterp)
  *     doesn't need to look it up in the search paths.
  *     The caller also checks whether pathname is already loaded or not,
  *     so this function doesn't need to worry about duplicate loads.
- *
  *     This function should have the semantics equivalent to the
  *     RTLD_NOW|RTLD_GLOBAL of dlopen().
  *
@@ -747,7 +868,7 @@ ScmObj Scm_DynLoad(ScmString *filename, ScmObj initfn, int export_)
  *   There may be a case that the feature dependency forms a loop because
  *   of bug.  An error should be signaled in such a case, rather than going
  *   to deadlock.   So, when the calling thread finds the required feature
- *   is in the ldinfo.providing alist, it checks the waiting chaing of
+ *   is in the ldinfo.providing alist, it checks the waiting chain of
  *   features, and no threads are waiting for a feature being provided by
  *   the calling thread.
  */
@@ -758,7 +879,7 @@ ScmObj Scm_Require(ScmObj feature)
     ScmVM *vm = Scm_VM();
     ScmObj provided, providing, p, q;
     int loop = FALSE;
-    
+
     if (!SCM_STRINGP(feature)) {
         Scm_Error("require: string expected, but got %S\n", feature);
     }
@@ -804,7 +925,7 @@ ScmObj Scm_Require(ScmObj feature)
     if (!SCM_FALSEP(provided)) return SCM_TRUE;
     SCM_UNWIND_PROTECT {
         filename = Scm_StringAppendC(SCM_STRING(feature), ".scm", 4, 4);
-        Scm_Load(Scm_GetStringConst(SCM_STRING(filename)), TRUE);
+        Scm_Load(Scm_GetStringConst(SCM_STRING(filename)), 0);
     } SCM_WHEN_ERROR {
         (void)SCM_INTERNAL_MUTEX_LOCK(ldinfo.prov_mutex);
         ldinfo.providing = Scm_AssocDeleteX(feature, ldinfo.providing, SCM_CMP_EQUAL);
@@ -917,6 +1038,8 @@ void Scm_DefineAutoload(ScmModule *where,
 ScmObj Scm_LoadAutoload(ScmAutoload *adata)
 {
     int error = FALSE;
+    ScmModule *prev_module;
+    ScmVM *vm = Scm_VM();
     
     /* check if some other thread already loaded this before attempt to lock */
     if (adata->loaded) {
@@ -928,14 +1051,14 @@ ScmObj Scm_LoadAutoload(ScmAutoload *adata)
     do {
         if (adata->loaded) break;
         if (adata->locker == NULL) {
-            adata->locker = Scm_VM(); /* take me */
-        } else if (adata->locker == Scm_VM()) {
+            adata->locker = vm;
+        } else if (adata->locker == vm) {
             /* bad circular dependency */
             error = TRUE;
         } else if (adata->locker->state == SCM_VM_TERMINATED) {
             /* the loading thread have died prematurely.
                let's take over the task. */
-            adata->locker = Scm_VM();
+            adata->locker = vm;
         } else {
             (void)SCM_INTERNAL_COND_WAIT(adata->cv, adata->mutex);
             continue;
@@ -954,20 +1077,24 @@ ScmObj Scm_LoadAutoload(ScmAutoload *adata)
                   adata->module, adata->name);
     }
 
+    prev_module = vm->module;
     SCM_UNWIND_PROTECT {
+        vm->module = adata->module;
         Scm_Require(SCM_OBJ(adata->path));
+        vm->module = prev_module;
     
         if (adata->import_from) {
             /* autoloaded file defines import_from module.  we need to
                import the binding individually. */
-            ScmObj m = Scm_FindModule(adata->import_from, FALSE);
+            ScmModule *m = Scm_FindModule(adata->import_from,
+                                          SCM_FIND_MODULE_QUIET);
             ScmGloc *f, *g;
-            if (!SCM_MODULEP(m)) {
+            if (m == NULL) {
                 Scm_Error("Trying to autoload module %S from file %S, but the file doesn't define such a module",
                           adata->import_from, adata->path);
             }
-            f = Scm_FindBinding(SCM_MODULE(m), adata->name, FALSE);
-            g = Scm_FindBinding(adata->module, adata->name, FALSE);
+            f = Scm_FindBinding(SCM_MODULE(m), adata->name, 0);
+            g = Scm_FindBinding(adata->module, adata->name, 0);
             SCM_ASSERT(f != NULL);
             SCM_ASSERT(g != NULL);
             adata->value = SCM_GLOC_GET(f);
@@ -979,7 +1106,7 @@ ScmObj Scm_LoadAutoload(ScmAutoload *adata)
         } else {
             /* Normal import.  The binding must have been inserted to
                adata->module */
-            ScmGloc *g = Scm_FindBinding(adata->module, adata->name, FALSE);
+            ScmGloc *g = Scm_FindBinding(adata->module, adata->name, 0);
             SCM_ASSERT(g != NULL);
             adata->value = SCM_GLOC_GET(g);
             if (SCM_UNBOUNDP(adata->value) || SCM_AUTOLOADP(adata->value)) {
@@ -989,6 +1116,7 @@ ScmObj Scm_LoadAutoload(ScmAutoload *adata)
         }
     } SCM_WHEN_ERROR {
         adata->locker = NULL;
+        vm->module = prev_module;
         SCM_INTERNAL_COND_SIGNAL(adata->cv);
         SCM_NEXT_HANDLER;
     } SCM_END_PROTECT;
@@ -999,55 +1127,6 @@ ScmObj Scm_LoadAutoload(ScmAutoload *adata)
     return adata->value;
 }
 
-/*------------------------------------------------------------------
- * Windows tricks
- */
-#ifdef __MINGW32__
-
-ScmObj get_install_dir(void)
-{
-    static ScmObj dir = SCM_FALSE;
-    if (SCM_FALSEP(dir)) {
-	HMODULE mod;
-	DWORD r;
-	char path[MAX_PATH];
-
-	mod = GetModuleHandle("libgauche.dll");
-	if (mod == NULL) {
-	    Scm_Error("GetModuleHandle failed");
-	}
-	r = GetModuleFileName(mod, path, MAX_PATH);
-	if (r == 0) {
-	    Scm_Error("GetModuleFileName failed");
-	}
-	/* remove \libgauche.dll */
-	if (!PathRemoveFileSpec(path)) {
-	    Scm_Error("PathRemoveFileSpec failed on %s", path);
-	}
-	/* remobe \bin */
-	if (!PathRemoveFileSpec(path)) {
-	    Scm_Error("PathRemoveFileSpec failed on %s", path);
-	}
-	dir = SCM_MAKE_STR_COPYING(path);
-    }
-    return dir;
-}
-
-ScmObj get_archdir(void)
-{
-    return Scm_StringAppendC(SCM_STRING(get_install_dir()),
-			     "\\lib\\gauche\\"GAUCHE_VERSION"\\"GAUCHE_ARCH,
-			     -1, -1);
-}
-
-ScmObj get_scmdir(void)
-{
-    return Scm_StringAppendC(SCM_STRING(get_install_dir()),
-			     "\\share\\gauche\\"GAUCHE_VERSION"\\lib",
-			     -1, -1);
-}
-#endif /*__MINGW32__*/
-
 /*------------------------------------------------------------------
  * Initialization
  */
@@ -1060,21 +1139,13 @@ void Scm__InitLoad(void)
 
     init_load_path = t = SCM_NIL;
     SCM_APPEND(init_load_path, t, break_env_paths("GAUCHE_LOAD_PATH"));
-#ifndef __MINGW32__
-    SCM_APPEND1(init_load_path, t, SCM_MAKE_STR(GAUCHE_SITE_LIB_DIR));
-    SCM_APPEND1(init_load_path, t, SCM_MAKE_STR(GAUCHE_LIB_DIR));
-#else  /*__MINGW32__*/
-    SCM_APPEND1(init_load_path, t, get_scmdir());
-#endif /*__MINGW32__*/
+    SCM_APPEND1(init_load_path, t, Scm_SiteLibraryDirectory());
+    SCM_APPEND1(init_load_path, t, Scm_LibraryDirectory());
 
     init_dynload_path = t = SCM_NIL;
     SCM_APPEND(init_dynload_path, t, break_env_paths("GAUCHE_DYNLOAD_PATH"));
-#ifndef __MINGW32__
-    SCM_APPEND1(init_dynload_path, t, SCM_MAKE_STR(GAUCHE_SITE_ARCH_DIR));
-    SCM_APPEND1(init_dynload_path, t, SCM_MAKE_STR(GAUCHE_ARCH_DIR));
-#else  /*__MINGW32__*/
-    SCM_APPEND1(init_dynload_path, t, get_archdir());
-#endif /*__MINGW32__*/
+    SCM_APPEND1(init_dynload_path, t, Scm_SiteArchitectureDirectory());
+    SCM_APPEND1(init_dynload_path, t, Scm_ArchitectureDirectory());
 
     init_load_suffixes = t = SCM_NIL;
     SCM_APPEND1(init_load_suffixes, t, SCM_MAKE_STR(LOAD_SUFFIX));
@@ -1103,6 +1174,7 @@ void Scm__InitLoad(void)
     key_error_if_not_found = SCM_MAKE_KEYWORD("error-if-not-found");
     key_environment = SCM_MAKE_KEYWORD("environment");
     key_macro = SCM_MAKE_KEYWORD("macro");
+    key_ignore_coding = SCM_MAKE_KEYWORD("ignore-coding");
     
     SCM_DEFINE(m, "load-from-port", SCM_OBJ(&load_from_port_STUB));
     SCM_DEFINE(m, "load", SCM_OBJ(&load_STUB));
@@ -1115,11 +1187,13 @@ void Scm__InitLoad(void)
     DEF(ldinfo.load_suffixes_rec, SCM_SYM_LOAD_SUFFIXES, init_load_suffixes);
     DEF(ldinfo.cond_features_rec, SCM_SYM_COND_FEATURES, init_cond_features);
 
-    ldinfo.provided = SCM_LIST4(SCM_MAKE_STR("srfi-6"), /* string ports (builtin) */
-                                SCM_MAKE_STR("srfi-8"), /* receive (builtin) */
-                                SCM_MAKE_STR("srfi-10"), /* #, (builtin) */
-                                SCM_MAKE_STR("srfi-17")  /* set! (builtin) */
-        );
+    ldinfo.provided =
+        SCM_LIST5(SCM_MAKE_STR("srfi-2"), /* and-let* */
+                  SCM_MAKE_STR("srfi-6"), /* string ports (builtin) */
+                  SCM_MAKE_STR("srfi-8"), /* receive (builtin) */
+                  SCM_MAKE_STR("srfi-10"), /* #, (builtin) */
+                  SCM_MAKE_STR("srfi-17")  /* set! (builtin) */
+            );
     ldinfo.providing = SCM_NIL;
     ldinfo.waiting = SCM_NIL;
     ldinfo.dso_suffixes = SCM_LIST2(SCM_MAKE_STR(".la"),
