@@ -1,7 +1,7 @@
 /*
  * number.c - numeric functions
  *
- *   Copyright (c) 2000-2004 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2005 Shiro Kawai, All rights reserved.
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: number.c,v 1.111 2004/06/27 23:27:28 shirok Exp $
+ *  $Id: number.c,v 1.124 2005/11/10 20:42:32 shirok Exp $
  */
 
 #include <math.h>
@@ -139,7 +139,7 @@ ScmObj Scm_MakeFlonum(double d)
 
 ScmObj Scm_MakeFlonumToNumber(double d, int exact)
 {
-    if (exact) {
+    if (exact && !SCM_IS_INF(d)) {
         /* see if d can be demoted to integer */
         double i, f;
         f = modf(d, &i);
@@ -170,6 +170,13 @@ ScmObj Scm_MakeFlonumToNumber(double d, int exact)
 union ieee_double {
     double d;
     struct {
+#ifdef DOUBLE_ARMENDIAN
+        /* ARM's mixed endian.  TODO: what if we have LP64 ARM? */
+        unsigned long mant0:20;
+        unsigned int exp:11;
+        unsigned int sign:1;
+        unsigned long mant1:32;
+#else  /*!DOUBLE_ARMENDIAN*/
 #ifdef WORDS_BIGENDIAN
 #if SIZEOF_LONG >= 8
         unsigned int sign:1;
@@ -193,6 +200,7 @@ union ieee_double {
         unsigned int  sign:1;
 #endif /*SIZEOF_LONG < 8*/
 #endif /*!WORDS_BIGENDIAN*/
+#endif /*!DOUBLE_ARMENDIAN*/
     } components;
 };
 
@@ -323,92 +331,141 @@ ScmObj Scm_MakeIntegerU(u_long i)
 }
 
 /* Convert scheme integer to C integer */
-long Scm_GetIntegerClamp(ScmObj obj, int clamphi, int clamplo)
+long Scm_GetIntegerClamp(ScmObj obj, int clamp, int *oor)
 {
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (SCM_INTP(obj)) return SCM_INT_VALUE(obj);
     else if (SCM_BIGNUMP(obj)) {
-        return Scm_BignumToSI(SCM_BIGNUM(obj), clamphi, clamplo);
+        return Scm_BignumToSI(SCM_BIGNUM(obj), clamp, oor);
     }
     else if (SCM_FLONUMP(obj)) {
         double v = SCM_FLONUM_VALUE(obj);
         if (v > (double)LONG_MAX) {
-            if (clamphi) return LONG_MAX;
+            if (clamp & SCM_CLAMP_HI) return LONG_MAX;
             else goto err;
         }
         if (v < (double)LONG_MIN) {
-            if (clamplo) return LONG_MIN;
+            if (clamp & SCM_CLAMP_LO) return LONG_MIN;
             else goto err;
         }
         return (long)v;
     }
   err:
-    Scm_Error("argument out of range: %S", obj);
-    return 0; /* dummy */
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) {
+        *oor = TRUE;
+    } else {
+        Scm_Error("argument out of range: %S", obj);
+    }
+    return 0;
 }
 
-u_long Scm_GetIntegerUClamp(ScmObj obj, int clamphi, int clamplo)
+u_long Scm_GetIntegerUClamp(ScmObj obj, int clamp, int *oor)
 {
-    if (SCM_INTP(obj) && SCM_INT_VALUE(obj) >= 0) return SCM_INT_VALUE(obj);
-    else if (SCM_BIGNUMP(obj)) {
-        return Scm_BignumToUI(SCM_BIGNUM(obj), clamphi, clamplo);
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
+    if (SCM_INTP(obj)) {
+        if (SCM_INT_VALUE(obj) < 0) {
+            if (clamp & SCM_CLAMP_LO) return 0;
+            else goto err;
+        }
+        return SCM_INT_VALUE(obj);
+    } else if (SCM_BIGNUMP(obj)) {
+        return Scm_BignumToUI(SCM_BIGNUM(obj), clamp, oor);
     }
     else if (SCM_FLONUMP(obj)) {
         double v = SCM_FLONUM_VALUE(obj);
         if (v > (double)ULONG_MAX) {
-            if (clamphi) return ULONG_MAX;
+            if (clamp & SCM_CLAMP_HI) return ULONG_MAX;
             else goto err;
         }
         if (v < 0.0) {
-            if (clamplo) return 0;
+            if (clamp & SCM_CLAMP_LO) return 0;
             else goto err;
         }
         return (u_long)v;
     }
   err:
-    Scm_Error("argument out of range: %S", obj);
-    return 0; /* dummy */
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) {
+        *oor = TRUE;
+    } else {
+        Scm_Error("argument out of range: %S", obj);
+    }
+    return 0;
 }
 
 /* 32bit integer specific */
-ScmInt32 Scm_GetInteger32Clamp(ScmObj obj, int clamphi, int clamplo)
+ScmInt32 Scm_GetInteger32Clamp(ScmObj obj, int clamp, int *oor)
 {
 #if SIZEOF_LONG == 4
-    return (ScmInt32)Scm_GetIntegerClamp(obj, clamphi, clamplo);
+    return (ScmInt32)Scm_GetIntegerClamp(obj, clamp, oor);
 #else  /* SIZEOF_LONG >= 8 */
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
+    /* NB: we denote the constant directly here.  (1L<<31) fails on
+       Alpha machines, since the compiler somehow calculates the constant
+       in 32bit integer even it has 'L'.  We have to write (1LL<<31), but
+       I'm afraid that it's not portable. */
     if (SCM_INTP(obj)) {
         long r = SCM_INT_VALUE(obj);
-        if (r < -(1L<<31)) {
-            if (clamplo) return -(1L<<31);
-            Scm_Error("argument out of range: %S", obj);
+        if (r < -0x80000000L) {
+            if (clamp & SCM_CLAMP_LO) return -0x80000000L;
+            goto err;
         }
-        if (r >= (1L<<31)) {
-            if (clamphi) return (1L<<31)-1;
-            Scm_Error("argument out of range: %S", obj);
+        if (r > 0x7fffffffL) {
+            if (clamp & SCM_CLAMP_HI) return 0x7fffffffL;
+            goto err;
+        }
+        return r;
+    } else if (SCM_BIGNUMP(obj)) {
+        if (SCM_BIGNUM_SIGN(obj) < 0) {
+            if (clamp & SCM_CLAMP_LO) return -0x80000000L;
+            goto err;
+        } else {
+            if (clamp & SCM_CLAMP_HI) return 0x7fffffffL;
+            goto err;
         }
     }
-    /*FALLTHROUGH*/
-    return (ScmInt32)Scm_GetIntegerClamp(obj, clamphi, clamplo);
+  err:
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) {
+        *oor = TRUE;
+    } else {
+        Scm_Error("argument out of range: %S", obj);
+    }
+    return 0;
 #endif /* SIZEOF_LONG >= 8 */
 }
 
-ScmUInt32 Scm_GetIntegerU32Clamp(ScmObj obj, int clamphi, int clamplo)
+ScmUInt32 Scm_GetIntegerU32Clamp(ScmObj obj, int clamp, int *oor)
 {
 #if SIZEOF_LONG == 4
-    return (ScmUInt32)Scm_GetIntegerUClamp(obj, clamphi, clamplo);
+    return (ScmUInt32)Scm_GetIntegerUClamp(obj, clamp, oor);
 #else  /* SIZEOF_LONG >= 8 */
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (SCM_INTP(obj)) {
         long r = SCM_INT_VALUE(obj);
         if (r < 0) {
-            if (clamplo) return 0;
-            Scm_Error("argument out of range: %S", obj);
+            if (clamp & SCM_CLAMP_LO) return 0;
+            goto err;
         }
-        if (r >= (1L<<32)) {
-            if (clamphi) return (1L<<32)-1;
-            Scm_Error("argument out of range: %S", obj);
+        if (r > 0xffffffffUL) {
+            if (clamp & SCM_CLAMP_HI) return 0xffffffffUL;
+            goto err;
+        }
+        return r;
+    } else if (SCM_BIGNUMP(obj)) {
+        if (SCM_BIGNUM_SIGN(obj) < 0) {
+            if (clamp & SCM_CLAMP_LO) return 0;
+            goto err;
+        } else {
+            if (clamp & SCM_CLAMP_HI) return 0xffffffffUL;
+            goto err;
         }
     }
-    /*FALLTHROUGH*/
-    return (ScmUInt32)Scm_GetIntegerUClamp(obj, clamphi, clamplo);
+  err:
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) {
+        *oor = TRUE;
+    } else {
+        Scm_Error("argument out of range: %S", obj);
+    }
+    return 0;
 #endif /* SIZEOF_LONG >= 8 */
 }
 
@@ -420,7 +477,6 @@ ScmObj Scm_MakeInteger64(ScmInt64 i)
 #if SCM_EMULATE_INT64
     u_long val[2];
     if (i.hi == 0) return Scm_MakeInteger(i.lo);
-    if (i.hi == ULONG_MAX) return Scm_MakeInteger((long)i.lo);
     val[0] = i.lo;
     val[1] = i.hi;
     return Scm_MakeBignumFromUIArray(0, val, 2); /* bignum checks sign */
@@ -428,9 +484,8 @@ ScmObj Scm_MakeInteger64(ScmInt64 i)
     u_long val[2];
     val[0] = (uint64_t)i & ULONG_MAX;
     val[1] = (uint64_t)i >> 32;
-    if (val[1] == 0) return Scm_MakeInteger(val[0]);
-    if (val[1] == ULONG_MAX) return Scm_MakeInteger((long)val[0]);
-    return Scm_MakeBignumFromUIArray(0, val, 2);
+    if (val[1] == 0 && val[0] <= LONG_MAX) return Scm_MakeInteger(val[0]);
+    return Scm_NormalizeBignum(SCM_BIGNUM(Scm_MakeBignumFromUIArray(0, val, 2)));
 #endif
 }
 
@@ -451,11 +506,11 @@ ScmObj Scm_MakeIntegerU64(ScmUInt64 i)
 #endif
 }
 
-ScmInt64 Scm_GetInteger64Clamp(ScmObj obj, int clamphi, int clamplo)
+ScmInt64 Scm_GetInteger64Clamp(ScmObj obj, int clamp, int *oor)
 {
 #if SCM_EMULATE_INT64
     ScmInt64 r = {0, 0};
-
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (SCM_INTP(obj)) {
         long v = SCM_INT_VALUE(obj);
         r.lo = v;
@@ -463,29 +518,28 @@ ScmInt64 Scm_GetInteger64Clamp(ScmObj obj, int clamphi, int clamplo)
         return r;
     }
     if (SCM_BIGNUMP(obj)) {
-        return Scm_BignumToSI64(SCM_BIGNUM(obj), clamphi, clamplo);
+        return Scm_BignumToSI64(SCM_BIGNUM(obj), clamp, oor);
     }
     if (SCM_FLONUMP(obj)) {
         if (Scm_NumCmp(obj, SCM_2_63) >= 0) {
-            if (!clamphi) goto err;
+            if (!(clamp&SCM_CLAMP_HI)) goto err;
             SCM_SET_INT64_MAX(r);
             return r;
         } else if (Scm_NumCmp(obj, SCM_MINUS_2_63) < 0) {
-            if (!clamplo) goto err;
+            if (!(clamp&SCM_CLAMP_LO)) goto err;
             SCM_SET_INT64_MIN(r);
             return r;
         } else {
             ScmObj b = Scm_MakeBignumFromDouble(SCM_FLONUM_VALUE(obj));
-            return Scm_BignumToSI64(SCM_BIGNUM(b), clamphi, clamplo);
+            return Scm_BignumToSI64(SCM_BIGNUM(b), clamp, oor);
         }
     }
-  err:
-    Scm_Error("argument out of range: %S", obj);
-    return r; /* dummy */
 #else /*!SCM_EMULATE_INT64*/
+    ScmInt64 r = 0;
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (SCM_INTP(obj)) return (ScmInt64)SCM_INT_VALUE(obj);
     if (SCM_BIGNUMP(obj)) {
-        return Scm_BignumToSI64(SCM_BIGNUM(obj), clamphi, clamplo);
+        return Scm_BignumToSI64(SCM_BIGNUM(obj), clamp, oor);
     }
     if (SCM_FLONUMP(obj)) {
         int64_t maxval, minval;
@@ -495,87 +549,94 @@ ScmInt64 Scm_GetInteger64Clamp(ScmObj obj, int clamphi, int clamplo)
         SCM_SET_INT64_MIN(minval);
         v = SCM_FLONUM_VALUE(obj);
         if (v > (double)maxval) {
-            if (!clamphi) goto err;
+            if (!(clamp&SCM_CLAMP_HI)) goto err;
             return maxval;
         } else if (v < (double)minval) {
-            if (!clamplo) goto err;
+            if (!(clamp&SCM_CLAMP_LO)) goto err;
             return minval;
         } else {
             return (long)v;
         }
     }
-  err:
-    Scm_Error("argument out of range: %S", obj);
-    return 0; /* dummy */
 #endif /*!SCM_EMULATE_INT64*/
+  err:
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) {
+        *oor = TRUE;
+    } else {
+        Scm_Error("argument out of range: %S", obj);
+    }
+    return r;
 }
                                
-ScmUInt64 Scm_GetIntegerU64Clamp(ScmObj obj, int clamphi, int clamplo)
+ScmUInt64 Scm_GetIntegerU64Clamp(ScmObj obj, int clamp, int *oor)
 {
 #if SCM_EMULATE_INT64
     ScmUInt64 r = {0, 0};
-
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (SCM_INTP(obj)) {
         long v = SCM_INT_VALUE(obj);
         if (v < 0) {
-            if (!clamplo) goto err;
+            if (!(clamp&SCM_CLAMP_LO)) goto err;
         } else {
             r.lo = v;
         }
         return r;
     }
     if (SCM_BIGNUMP(obj)) {
-        return Scm_BignumToUI64(SCM_BIGNUM(obj), clamphi, clamplo);
+        return Scm_BignumToUI64(SCM_BIGNUM(obj), clamp, oor);
     }
     if (SCM_FLONUMP(obj)) {
         if (Scm_NumCmp(obj, SCM_2_64) >= 0) {
-            if (!clamphi) goto err;
+            if (!(clamp&SCM_CLAMP_HI)) goto err;
             SCM_SET_UINT64_MAX(r);
             return r;
         } else if (SCM_FLONUM_VALUE(obj) < 0) {
-            if (!clamplo) goto err;
+            if (!(clamp&SCM_CLAMP_LO)) goto err;
             return r;
         } else {
             ScmObj b = Scm_MakeBignumFromDouble(SCM_FLONUM_VALUE(obj));
-            return Scm_BignumToUI64(SCM_BIGNUM(b), clamphi, clamplo);
+            return Scm_BignumToUI64(SCM_BIGNUM(b), clamp, oor);
         }
     }
-  err:
-    Scm_Error("argument out of range: %S", obj);
-    return r; /* dummy */
-#else /*SCM_EMULATE_INT64*/
+#else /*!SCM_EMULATE_INT64*/
+    ScmInt64 r = 0;
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (SCM_INTP(obj)) {
         long v = SCM_INT_VALUE(obj);
         if (v < 0) {
-            if (!clamplo) goto err;
+            if (!(clamp&SCM_CLAMP_LO)) goto err;
             return 0;
         } else {
             return (ScmUInt64)v;
         }
     }
     if (SCM_BIGNUMP(obj)) {
-        return Scm_BignumToUI64(SCM_BIGNUM(obj), clamphi, clamplo);
+        return Scm_BignumToUI64(SCM_BIGNUM(obj), clamp, oor);
     }
     if (SCM_FLONUMP(obj)) {
         double v = SCM_FLONUM_VALUE(obj);
         uint64_t maxval;
 
         if (v < 0) {
-            if (!clamplo) goto err;
+            if (!(clamp&SCM_CLAMP_LO)) goto err;
             return 0;
         }
         SCM_SET_UINT64_MAX(maxval);
         if (v > (double)maxval) {
-            if (!clamphi) goto err;
+            if (!(clamp&SCM_CLAMP_HI)) goto err;
             return maxval;
         } else {
             return (uint32_t)v;
         }
     }
-  err:
-    Scm_Error("argument out of range: %S", obj);
-    return 0; /* dummy */
 #endif
+  err:
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) {
+        *oor = TRUE;
+    } else {
+        Scm_Error("argument out of range: %S", obj);
+    }
+    return r;
 }
                                
 #endif /* SIZEOF_LONG == 4 */
@@ -701,21 +762,17 @@ ScmObj Scm_Reciprocal(ScmObj obj)
 {
     if (SCM_INTP(obj)) {
         long val = SCM_INT_VALUE(obj);
-        if (val == 0) Scm_Error("divide by zero");
         obj = Scm_MakeFlonum(1.0/(double)val);
     } else if (SCM_BIGNUMP(obj)) {
         double val = Scm_BignumToDouble(SCM_BIGNUM(obj));
-        if (val == 0.0) Scm_Error("divide by zero");
         obj = Scm_MakeFlonum(1.0/val);
     } else if (SCM_FLONUMP(obj)) {
         double val = SCM_FLONUM_VALUE(obj);
-        if (val == 0.0) Scm_Error("divide by zero");
         obj = Scm_MakeFlonum(1.0/val);
     } else if (SCM_COMPLEXP(obj)) {
         double r = SCM_COMPLEX_REAL(obj), r1;
         double i = SCM_COMPLEX_IMAG(obj), i1;
         double d;
-        if (r == 0.0 && i == 0.0) Scm_Error("divide by zero");
         d = r*r + i*i;
         r1 = r/d;
         i1 = -i/d;
@@ -792,6 +849,7 @@ ScmObj Scm_PromoteToComplex(ScmObj obj)
         return Scm_MakeComplex(Scm_BignumToDouble(SCM_BIGNUM(obj)), 0.0);
     if (SCM_FLONUMP(obj))
         return Scm_MakeComplex(SCM_FLONUM_VALUE(obj), 0.0);
+    if (SCM_COMPLEXP(obj)) return obj;
     Scm_Panic("Scm_PromoteToComplex: can't be here");
     return SCM_UNDEFINED;       /* dummy */
 }
@@ -1152,9 +1210,20 @@ ScmObj Scm_Divide(ScmObj arg0, ScmObj arg1, ScmObj args)
            inexact number */
         if (SCM_INTP(arg1)) {
             long rem;
-            ScmObj div = Scm_BignumDivSI(SCM_BIGNUM(arg0),
-                                         SCM_INT_VALUE(arg1),
-                                         &rem);
+            ScmObj div;
+
+            if (SCM_EQ(arg1, SCM_MAKE_INT(0))) {
+                if (SCM_BIGNUM_SIGN(arg0) > 0) {
+                    result_real = 1.0 / 0.0; /* +inf */
+                } else {
+                    result_real = -1.0 / 0.0; /* -inf */
+                }
+                exact = FALSE;
+                goto DO_FLONUM;
+            }
+            div = Scm_BignumDivSI(SCM_BIGNUM(arg0),
+                                  SCM_INT_VALUE(arg1),
+                                  &rem);
             if (rem != 0) {
                 result_real = Scm_BignumToDouble(SCM_BIGNUM(arg0));
                 exact = FALSE;
@@ -1204,7 +1273,6 @@ ScmObj Scm_Divide(ScmObj arg0, ScmObj arg1, ScmObj args)
                                     Scm_MakeFlonumToNumber(result_real, exact),
                                     arg1, args);
             }
-            if (div_real == 0) Scm_Error("divide by zero");
             result_real /= div_real;
             if (SCM_NULLP(args))
                 return Scm_MakeFlonumToNumber(result_real, exact);
@@ -1235,7 +1303,6 @@ ScmObj Scm_Divide(ScmObj arg0, ScmObj arg1, ScmObj args)
                                     arg1, args);
             }
             d = div_real*div_real + div_imag*div_imag;
-            if (d == 0.0) Scm_Error("divide by zero");
             r = (result_real*div_real + result_imag*div_imag)/d;
             i = (result_imag*div_real - result_real*div_imag)/d;
             result_real = r;
@@ -1325,7 +1392,7 @@ ScmObj Scm_Quotient(ScmObj x, ScmObj y, ScmObj *rem)
         goto BADARG;
     }
   DIVBYZERO:
-    Scm_Error("divide by zero");
+    Scm_Error("attempt to calculate a quotient by zero");
   BADARGY:
     x = y;
   BADARG:
@@ -1431,7 +1498,7 @@ ScmObj Scm_Modulo(ScmObj x, ScmObj y, int remp)
         goto BADARG;
     }
   DIVBYZERO:
-    Scm_Error("divide by zero");
+    Scm_Error("attempt to take a modulo or remainder by zero");
   BADARGY:
     x = y;
   BADARG:
@@ -1446,7 +1513,8 @@ ScmObj Scm_Modulo(ScmObj x, ScmObj y, int remp)
 /* Integer power of 10.  It is extensively used during string->number
    and number->string operations.
    IEXPT10_TABLESIZ is ceil(-log10(ldexp(1.0, -1022-52))) + 2 */
-#define IEXPT10_TABLESIZ  326
+/* NB: actually we need more margin here to handle denormalized numbers. */
+#define IEXPT10_TABLESIZ  341
 static ScmObj iexpt10_n[IEXPT10_TABLESIZ] = { NULL };
 static int    iexpt10_initialized = FALSE;
 
@@ -1655,12 +1723,12 @@ static double roundeven(double v)
     if (v > 0.0) {
         if (frac > 0.5) r += 1.0;
         else if (frac == 0.5) {
-            if (r/2.0 != 0.0) r += 1.0;
+            if (fmod(r,2.0) != 0.0) r += 1.0;
         }
     } else {
         if (frac < -0.5) r -= 1.0;
-        else if (frac == 0.5) {
-            if (r/2.0 != 0.0) r -= 1.0;
+        else if (frac == -0.5) {
+            if (fmod(r,2.0) != 0.0) r -= 1.0;
         }
     }
     return r;
@@ -1869,15 +1937,25 @@ static inline int numcmp3(ScmObj x, ScmObj d, ScmObj y)
 
 static void double_print(char *buf, int buflen, double val, int plus_sign)
 {
-    if (val < 0.0) *buf++ = '-', buflen--;
-    else if (plus_sign) *buf++ = '+', buflen--;
-    if (SCM_IS_INF(val)) {
-        strcpy(buf, "#<inf>");
+    /* Handle a few special cases first.
+       The notation of infinity is provisional; see how srfi-70 becomes. */
+    if (val == 0.0) {
+        if (plus_sign) strcpy(buf, "+0.0");
+        else strcpy(buf, "0.0");
+        return;
+    } else if (SCM_IS_INF(val)) {
+        if (val < 0.0) strcpy(buf, "#i-1/0");
+        else if (plus_sign) strcpy(buf, "#i+1/0");
+        else strcpy(buf, "#i1/0");
+        return;
     } else if (SCM_IS_NAN(val)) {
         strcpy(buf, "#<nan>");
-    } else if (val == 0.0) {
-        strcpy(buf, "0.0");
-    } else {
+        return;
+    }
+    
+    if (val < 0.0) *buf++ = '-', buflen--;
+    else if (plus_sign) *buf++ = '+', buflen--;
+    {
         /* variable names follows Burger&Dybvig paper. mp, mm for m+, m-.
            note that m+ == m- for most cases, and m+ == 2*m- for the rest.
            so we calculate m+ from m- for each iteration, using the flag
@@ -2078,6 +2156,15 @@ ScmObj Scm_NumberToString(ScmObj obj, int radix, int use_upper)
     return r;
 }
 
+/* utility to expose Burger&Dybvig algorithm.  FLAGS is not used yet,
+   but reserved for future extension. */
+void Scm_PrintDouble(ScmPort *port, double d, int flags)
+{
+    char buf[FLT_BUF];
+    double_print(buf, FLT_BUF, d, FALSE);
+    Scm_Putz(buf, strlen(buf), port);
+}
+
 /*
  * Number Parser
  *
@@ -2377,8 +2464,11 @@ static ScmObj read_real(const char **strp, int *lenp,
             if (SCM_FALSEP(denom)) return SCM_FALSE;
             if (denom == SCM_MAKE_INT(0)) {
                 if (lensave > *lenp) {
-                    return numread_error("(zero in denominator of rational number)",
-                                         ctx);
+                    if (minusp) {
+                        return Scm_MakeFlonum(-1.0/0.0);
+                    } else {
+                        return Scm_MakeFlonum(1.0/0.0);
+                    }
                 } else {
                     return SCM_FALSE;
                 }
@@ -2604,12 +2694,13 @@ static ScmObj numread_error(const char *msg, struct numread_packet *context)
 
 ScmObj Scm_StringToNumber(ScmString *str, int radix, int strict)
 {
-    if (SCM_STRING_LENGTH(str) != SCM_STRING_SIZE(str)) {
+    u_int len, size;
+    const char *p = Scm_GetStringContent(str, &size, &len, NULL);
+    if (size != len) {
         /* This can't be a proper number. */
         return SCM_FALSE;
     } else {
-        return read_number(SCM_STRING_START(str), SCM_STRING_SIZE(str),
-                           radix, strict);
+        return read_number(p, size, radix, strict);
     }
 }
 
@@ -2640,6 +2731,7 @@ void Scm__InitNumber(void)
     
     SCM_2_63 = Scm_Ash(SCM_MAKE_INT(1), 63);
     SCM_2_64 = Scm_Ash(SCM_MAKE_INT(1), 64);
+    SCM_2_64_MINUS_1 = Scm_Subtract2(SCM_2_64, SCM_MAKE_INT(1));
     SCM_2_52 = Scm_Ash(SCM_MAKE_INT(1), 52);
     SCM_2_53 = Scm_Ash(SCM_MAKE_INT(1), 53);
     SCM_MINUS_2_63 = Scm_Negate(SCM_2_63);
