@@ -1,7 +1,7 @@
 /*
  * read.c - reader
  *
- *   Copyright (c) 2000-2004 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2005 Shiro Kawai, All rights reserved.
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: read.c,v 1.73 2004/05/21 03:43:29 shirok Exp $
+ *  $Id: read.c,v 1.83 2005/10/13 08:14:13 shirok Exp $
  */
 
 #include <stdio.h>
@@ -204,7 +204,7 @@ void Scm_ReadError(ScmPort *port, const char *msg, ...)
     va_end(ap);
 
     rerr = Scm_MakeReadError(Scm_GetOutputString(SCM_PORT(ostr)), port, line);
-    Scm_VMThrowException(rerr);
+    Scm_Raise(rerr);
 }
 
 /*----------------------------------------------------------------
@@ -363,7 +363,7 @@ static void read_nested_comment(ScmPort *port, ScmReadContext *ctx)
             continue;
         case EOF:
           eof:
-            Scm_Error("encountered EOF inside nested multi-line comment (comment begins at line %d)", line);
+            Scm_ReadError(port, "encountered EOF inside nested multi-line comment (comment begins at line %d)", line);
         default:
             break;
         }
@@ -384,9 +384,7 @@ static int skipws(ScmPort *port, ScmReadContext *ctx)
                 c = Scm_GetbUnsafe(port);
                 if (c == '\n') {
                     /* oops.  ugly. */
-                    if (SCM_PORT_TYPE(port) == SCM_PORT_FILE) {
-                        port->src.buf.line++;
-                    }
+                    port->line++;
                     break;
                 }
                 if (c == EOF) return EOF;
@@ -437,12 +435,6 @@ static ScmObj read_internal(ScmPort *port, ScmReadContext *ctx)
             case '[':
                 /* #[...] literal charset */
                 return read_charset(port);
-            case '"':
-                /* #"..." explicit incomplete string */
-                /* NB: this syntax will be taken by string interpolation
-                   in future.  Use #*"..." instead. */
-                Scm_ReadError(port, "syntax #\"...\" for incomplete string is obsoleted.  use #*\"...\" instead.");
-                return SCM_UNDEFINED;
             case ',':
                 /* #,(form) - SRFI-10 read-time macro */
                 return read_sharp_comma(port, ctx);
@@ -563,45 +555,30 @@ static ScmObj read_item(ScmPort *port, ScmReadContext *ctx)
 /* Internal read_list.  returns whether the list contains unresolved
    reference or not within the flag has_ref */
 static ScmObj read_list_int(ScmPort *port, ScmChar closer,
-                            ScmReadContext *ctx, int *has_ref)
+                            ScmReadContext *ctx, int *has_ref, int start_line)
 {
     ScmObj start = SCM_NIL, last = SCM_NIL, item;
     int c, dot_seen = FALSE, ref_seen = FALSE;
-    int line = -1;
 
-    if (ctx->flags & SCM_READ_SOURCE_INFO) line = Scm_PortLine(port);
-    
     for (;;) {
         c = skipws(port, ctx);
-        if (c == EOF) {
-            if (line >= 0) {
-                Scm_ReadError(port, "EOF inside a list (starting from line %d)", line);
-            } else {
-                Scm_ReadError(port, "EOF inside a list");
-            }
-        }
+        if (c == EOF) goto eoferr;
         if (c == closer) {
             *has_ref = !!ref_seen;
             return start;
         }
 
-        if (dot_seen) Scm_ReadError(port, "bad dot syntax");
+        if (dot_seen) goto baddot;
 
         if (c == '.') {
             int c2 = Scm_GetcUnsafe(port);
-            if (c2 == closer) { 
-                Scm_ReadError(port, "bad dot syntax");
+            if (c2 == closer) {
+                goto baddot;
             } else if (c2 == EOF) {
-                if (line >= 0) {
-                    Scm_ReadError(port, "EOF inside a list (starting from line %d)", line);
-                } else {
-                    Scm_ReadError(port, "EOF inside a list");
-                }
+                goto eoferr;
             } else if (isspace(c2)) {
                 /* dot pair at the end */
-                if (start == SCM_NIL) {
-                    Scm_ReadError(port, "bad dot syntax");
-                }
+                if (start == SCM_NIL) goto baddot;
                 item = read_item(port, ctx);
                 if (SCM_READ_REFERENCE_P(item)) ref_seen = TRUE;
                 SCM_SET_CDR(last, item);
@@ -617,19 +594,37 @@ static ScmObj read_list_int(ScmPort *port, ScmChar closer,
             if (SCM_READ_REFERENCE_P(item)) ref_seen = TRUE;
         }
         SCM_APPEND1(start, last, item);
-        if (start==last && (ctx->flags & SCM_READ_SOURCE_INFO) && line >= 0) {
-            /* add source information to the top of the list */
-            Scm_PairAttrSet(SCM_PAIR(start), SCM_SYM_SOURCE_INFO,
-                            SCM_LIST2(Scm_PortName(port),
-                                      SCM_MAKE_INT(line)));
-        }
     }
+  eoferr:
+    if (start_line >= 0) {
+        Scm_ReadError(port, "EOF inside a list (starting from line %d)",
+                      start_line);
+    } else {
+        Scm_ReadError(port, "EOF inside a list");
+    }
+  baddot:
+    Scm_ReadError(port, "bad dot syntax");
+    return SCM_NIL;             /* dummy */
 }
 
 static ScmObj read_list(ScmPort *port, ScmChar closer, ScmReadContext *ctx)
 {
     int has_ref;
-    ScmObj r = read_list_int(port, closer, ctx, &has_ref);
+    int line = -1;
+    ScmObj r;
+
+    if (ctx->flags & SCM_READ_SOURCE_INFO) line = Scm_PortLine(port);
+
+    r = read_list_int(port, closer, ctx, &has_ref, line);
+
+    if (SCM_PAIRP(r) && (ctx->flags & SCM_READ_SOURCE_INFO) && line >= 0) {
+        /* Swap the head of the list for an extended pair to record
+           source-code info.*/
+        r = Scm_ExtendedCons(SCM_CAR(r), SCM_CDR(r));
+        Scm_PairAttrSet(SCM_PAIR(r), SCM_SYM_SOURCE_INFO,
+                        SCM_LIST2(Scm_PortName(port), SCM_MAKE_INT(line)));
+    }
+
     if (has_ref) ref_push(ctx, r, SCM_FALSE);
     return r;
 }
@@ -637,8 +632,12 @@ static ScmObj read_list(ScmPort *port, ScmChar closer, ScmReadContext *ctx)
 static ScmObj read_vector(ScmPort *port, ScmChar closer, ScmReadContext *ctx)
 {
     int has_ref;
-    ScmObj r = read_list_int(port, closer, ctx, &has_ref);
-    r = Scm_ListToVector(r);
+    int line = -1;
+    ScmObj r;
+    
+    if (ctx->flags & SCM_READ_SOURCE_INFO) line = Scm_PortLine(port);
+    r = read_list_int(port, closer, ctx, &has_ref, line);
+    r = Scm_ListToVector(r, 0, -1);
     if (has_ref) ref_push(ctx, r, SCM_FALSE);
     return r;
 }
@@ -651,10 +650,12 @@ static ScmObj read_quoted(ScmPort *port, ScmObj quoter, ScmReadContext *ctx)
     if (ctx->flags & SCM_READ_SOURCE_INFO) line = Scm_PortLine(port);
     item = read_item(port, ctx);
     if (SCM_EOFP(item)) Scm_ReadError(port, "unterminated quote");
-    r = Scm_Cons(quoter, Scm_Cons(item, SCM_NIL));
     if (line >= 0) {
+        r = Scm_ExtendedCons(quoter, Scm_Cons(item, SCM_NIL));
         Scm_PairAttrSet(SCM_PAIR(r), SCM_SYM_SOURCE_INFO,
                         SCM_LIST2(Scm_PortName(port), SCM_MAKE_INT(line)));
+    } else {
+        r = Scm_Cons(quoter, Scm_Cons(item, SCM_NIL));
     }
     if (SCM_READ_REFERENCE_P(item)) ref_push(ctx, SCM_CDR(r), SCM_FALSE);
     return r;
@@ -691,8 +692,9 @@ static ScmChar read_string_xdigits(ScmPort *port, int ndigs, int key,
         Scm_DStringPutc(&ds, '\\');
         Scm_DStringPutc(&ds, key);
         for (i=0; i<nread; i++) Scm_DStringPutc(&ds, (unsigned char)buf[i]);
-        Scm_Error("Bad '\\%c' escape sequence in a string literal: %s",
-                  key, Scm_DStringGetz(&ds));
+        Scm_ReadError(port,
+                      "Bad '\\%c' escape sequence in a string literal: %s",
+                      key, Scm_DStringGetz(&ds));
     }
     return r;
 }
@@ -716,11 +718,9 @@ static ScmObj read_string(ScmPort *port, int incompletep,
         switch (c) {
           case EOF: goto eof_exit;
           case '"': {
-              ScmString *s = SCM_STRING(Scm_DStringGet(&ds));
-              if (incompletep) {
-                  s = SCM_STRING(Scm_StringCompleteToIncompleteX(s));
-              }
-              return Scm_StringMakeImmutable(s);
+              int flags = ((incompletep? SCM_STRING_INCOMPLETE : 0)
+                           | SCM_STRING_IMMUTABLE);
+              return Scm_DStringGet(&ds, flags);
           }
           case '\\': {
             int c1 = Scm_GetcUnsafe(port);
@@ -756,7 +756,8 @@ static ScmObj read_string(ScmPort *port, int incompletep,
         }
     }
  eof_exit:
-    Scm_ReadError(port, "EOF encountered in a string literal: %S", Scm_DStringGet(&ds));
+    Scm_ReadError(port, "EOF encountered in a string literal: %S",
+                  Scm_DStringGet(&ds, 0));
     /* NOTREACHED */
     return SCM_FALSE; 
 }
@@ -791,6 +792,7 @@ static ScmObj read_char(ScmPort *port, ScmReadContext *ctx)
     int c;
     ScmString *name;
     const char *cname;
+    u_int namelen, namesize;
     struct char_name *cntab = char_names;
     
     c = Scm_GetcUnsafe(port);
@@ -803,26 +805,26 @@ static ScmObj read_char(ScmPort *port, ScmReadContext *ctx)
     default:
         /* need to read word to see if it is a character name */
         name = SCM_STRING(read_word(port, c, ctx, TRUE));
-        if (SCM_STRING_LENGTH(name) == 1) {
+        cname = Scm_GetStringContent(name, &namesize, &namelen, NULL);
+        if (namelen == 1) {
             return SCM_MAKE_CHAR(c);
         }
-        cname = Scm_GetStringConst(name);
-        if (SCM_STRING_LENGTH(name) != SCM_STRING_SIZE(name)) {
+        if (namelen != namesize) {
             /* no character name contains multibyte chars */
             goto unknown;
         }
 
         /* handle #\x1f etc. */
         if (cname[0] == 'x' && isxdigit(cname[1])) {
-            int code = Scm_ReadXdigitsFromString(cname+1, SCM_STRING_SIZE(name)-1, NULL);
+            int code = Scm_ReadXdigitsFromString(cname+1, namesize-1, NULL);
             if (code < 0) goto unknown;
             return SCM_MAKE_CHAR(code);
         }
         /* handle #\uxxxx or #\uxxxxxxxx*/
         if ((cname[0] == 'u') && isxdigit(cname[1])) {
             int code;
-            if (SCM_STRING_SIZE(name) == 5 || SCM_STRING_SIZE(name) == 9) {
-                code = Scm_ReadXdigitsFromString(cname+1, SCM_STRING_SIZE(name)-1, NULL);
+            if (namesize == 5 || namesize == 9) {
+                code = Scm_ReadXdigitsFromString(cname+1, namesize-1, NULL);
                 if (code >= 0) return SCM_MAKE_CHAR(Scm_UcsToChar(code));
             }
             /* if we come here, it's an error. */
@@ -830,11 +832,11 @@ static ScmObj read_char(ScmPort *port, ScmReadContext *ctx)
         }
 
         while (cntab->name) {
-            if (strcmp(cntab->name, cname) == 0) return cntab->ch;
+            if (strncmp(cntab->name, cname, namesize) == 0) return cntab->ch;
             cntab++;
         }
       unknown:
-        Scm_ReadError(port, "Unknown character name: #\\%s", cname);
+        Scm_ReadError(port, "Unknown character name: #\\%A", name);
     }
     return SCM_UNDEFINED;       /* dummy */
 }
@@ -843,6 +845,11 @@ static ScmObj read_char(ScmPort *port, ScmReadContext *ctx)
  * Symbols and Numbers
  */
 
+/* Reads a sequence of word-constituent characters from PORT, and returns
+   ScmString.  INITIAL may be a readahead character, or SCM_CHAR_INVALID
+   if there's none.  TEMP_CASE_FOLD turns on case-fold mode regardless of
+   the read context setting.
+*/
 static ScmObj read_word(ScmPort *port, ScmChar initial, ScmReadContext *ctx,
                         int temp_case_fold)
 {
@@ -859,7 +866,7 @@ static ScmObj read_word(ScmPort *port, ScmChar initial, ScmReadContext *ctx,
         c = Scm_GetcUnsafe(port);
         if (c == EOF || !char_word_constituent(c)) {
             Scm_UngetcUnsafe(c, port); 
-            return Scm_DStringGet(&ds);
+            return Scm_DStringGet(&ds, 0);
         }
         if (case_fold && char_word_case_fold(c)) c = tolower(c);
         SCM_DSTRING_PUTC(&ds, c);
@@ -908,11 +915,13 @@ static ScmObj read_escaped_symbol(ScmPort *port, ScmChar delim)
         if (c == EOF) {
             goto err;
         } else if (c == delim) {
-            ScmString *s = SCM_STRING(Scm_DStringGet(&ds));
+            ScmString *s = SCM_STRING(Scm_DStringGet(&ds, 0));
             return Scm_Intern(s);
         } else if (c == '\\') {
             /* CL-style single escape */
             c = Scm_GetcUnsafe(port);
+            /* TODO: we should recognize \xNN, since the symbol writer
+               prints a symbol name in that syntax. */
             if (c == EOF) goto err;
             SCM_DSTRING_PUTC(&ds, c);
         } else {
@@ -953,7 +962,7 @@ static ScmObj read_regexp(ScmPort *port)
             c = Scm_GetcUnsafe(port);
             if (c == 'i') flags |= SCM_REGEXP_CASE_FOLD;
             else          Scm_UngetcUnsafe(c, port);
-            return Scm_RegComp(SCM_STRING(Scm_DStringGet(&ds)), flags);
+            return Scm_RegComp(SCM_STRING(Scm_DStringGet(&ds, 0)), flags);
         } else {
             SCM_DSTRING_PUTC(&ds, c);
         }
@@ -1008,7 +1017,8 @@ static ScmObj read_reference(ScmPort *port, ScmChar ch, ScmReadContext *ctx)
         ScmObj ref = Scm_MakeReadReference();
 
         if (ctx->table == NULL) {
-            ctx->table = SCM_HASHTABLE(Scm_MakeHashTable((ScmHashProc)SCM_HASH_EQV, NULL, 0));
+            ctx->table =
+                SCM_HASH_TABLE(Scm_MakeHashTableSimple(SCM_HASH_EQV, 0));
         }
         if (Scm_HashTableGet(ctx->table, Scm_MakeInteger(refnum)) != NULL) {
             Scm_ReadError(port, "duplicate back-reference number in #%d=", refnum);
@@ -1039,7 +1049,7 @@ ScmObj Scm_DefineReaderCtor(ScmObj symbol, ScmObj proc, ScmObj finisher)
 
 static ScmObj read_sharp_comma(ScmPort *port, ScmReadContext *ctx)
 {
-    int len, has_ref;
+    int len, has_ref, line=-1;
     ScmChar next;
     ScmObj form, r;
 
@@ -1048,8 +1058,10 @@ static ScmObj read_sharp_comma(ScmPort *port, ScmReadContext *ctx)
         Scm_ReadError(port, "bad #,-form: '(' should be followed, but got %C",
                       next);
     }
+
+    if (ctx->flags & SCM_READ_SOURCE_INFO) line = Scm_PortLine(port);
     
-    form = read_list_int(port, ')', ctx, &has_ref);
+    form = read_list_int(port, ')', ctx, &has_ref, line);
     len = Scm_Length(form);
     if (len <= 0) {
         Scm_ReadError(port, "bad #,-form: #,%S", form);
@@ -1135,7 +1147,7 @@ static ScmObj maybe_uvector(ScmPort *port, char ch, ScmReadContext *ctx)
            serialized in Scm_Require. */
         Scm_Require(SCM_MAKE_STR("gauche/uvector"));
         if (Scm_ReadUvectorHook == NULL)
-            Scm_Error("couldn't load srfi-4 module");
+            Scm_ReadError(port, "couldn't load srfi-4 module");
     }
     return Scm_ReadUvectorHook(port, tag, ctx);
 }
@@ -1146,8 +1158,8 @@ static ScmObj maybe_uvector(ScmPort *port, char ch, ScmReadContext *ctx)
 
 void Scm__InitRead(void)
 {
-    readCtorData.table = SCM_HASHTABLE(Scm_MakeHashTable(SCM_HASH_ADDRESS,
-                                                         NULL, 0));
+    readCtorData.table =
+        SCM_HASH_TABLE(Scm_MakeHashTableSimple(SCM_HASH_EQ, 0));
     (void)SCM_INTERNAL_MUTEX_INIT(readCtorData.mutex);
     Scm_DefineReaderCtor(SCM_SYM_DEFINE_READER_CTOR,
                          Scm_MakeSubr(reader_ctor, NULL, 2, 1,
